{****************************************************************************

                       Copyright (c) 1996 by Florian Klmpfl

 ****************************************************************************}

unit link;

{ linker of FPKPascal }

  interface

    uses
       strings,objects,cobjects,globals
{$ifdef i386}
       ,i386
{$else}
{$endif}
       ;


  implementation

    const
       seg_text = 4;
       seg_data = 6;
       seg_bss = 8;

    type
       plinksym = ^tlinksym;

       tlinksym = object
          name : pchar;
          addr : longint;
          left,right : plinksym;
          seg : byte;
          constructor init(n : pchar;a : longint;s : byte);
       end;

       texec = record
          magic : longint;
          textlen : longint;
          datalen : longint;
          bsslen : longint;
          symslen : longint;
          startaddr : longint;
          reloctext : longint;
          relocdata : longint;
       end;

       preloc = ^treloc;

       treloc = object
          reloc_symbol : plinksym;
          reloc_info : longint;
          next : preloc;
       end;

       pobjfile = ^tobjfile;

       tobjfile = object
          textsize,datasize,bsssize,bssstart,textstart,datastart,
          textrelocs_count,datarelocs_count : longint;
          textseg,dataseg : pbyte;
          datarelocs : preloc;
          strings : pchar;
          next : pobjfile;
       end;

       pdos_o_file = ^tdos_o_file;

       tdos_o_file = object(tobjfile)
          static_syms : plinksym;
          constructor load(var f : file);
       end;

    var
       textendaddr : longint;
       dataendaddr : longint;
       bssendaddr : longint;
       global_syms : plinksym;
       ofiles : pobjfile;

    constructor tlinksym.init(n : pchar;a : longint;s : byte);

      begin
         left:=nil;
         right:=nil;
         { owner is the o-file }
         name:=n;
         addr:=a;
         seg:=s;
{$ifdef DEBUG_LINK}
         writeln(name,' at ',addr,' in Segment ',seg);
{$endif}
      end;

    procedure insert(p : plinksym;var t : plinksym);

      procedure do_insert(var at : plinksym);

        var
           l : longint;

        begin
           if not(assigned(at)) then
             at:=p
           else
             begin
                l:=strcomp(at^.name,p^.name);
                if l=0 then
                  begin
                     exterror:=strnew(p^.name);
                     error(linker_dup_symbol);
                  end
                else if (l>0) then
                  do_insert(at^.left)
                else do_insert(at^.right);
             end;
        end;

      begin
         do_insert(t);
      end;

    function search(p : pchar;t : plinksym) : plinksym;

      function do_search(at : plinksym) : plinksym;

        var
           l : longint;

        begin
           if not(assigned(at)) then
             do_search:=nil
           else
             begin
                l:=strcomp(at^.name,p);
                if l=0 then
                  do_search:=at
                else if (l>0) then
                  do_search:=do_search(at^.left)
                else
                  do_search:=do_search(at^.right);
             end;
        end;

      begin
         search:=do_search(t);
      end;

    constructor tdos_o_file.load(var f : file);

      type
         tsym = record
            name : longint;
            _type : byte;
            other : byte;
            desc : word;
            value : longint;
         end;

         tsymarray = array[0..65530 div 12] of tsym;

         psymarray = ^tsymarray;

         tlreloc = record
            addr : longint;
            ext : longint;
         end;

         trelocarray = array[0..65500 div 8] of tlreloc;

         prelocarray = ^trelocarray;

      var
         header : texec;
         syms : psymarray;
         _textrelocs,_datarelocs : prelocarray;
         l : longint;
         hs : plinksym;

      begin
         static_syms:=nil;
         seek(f,0);
         blockread(f,header,sizeof(header));
         if ioresult<>0 then
           fatalerror(linker_file_read_error);

         { calculate the segment start addresses }
         textstart:=textendaddr;
         datastart:=dataendaddr;
         bssstart:=bssendaddr;

         { load text segment }
         textsize:=header.textlen;
         getmem(textseg,textsize);
         blockread(f,textseg^,textsize);
         if ioresult<>0 then
           fatalerror(linker_file_read_error);
         { ... and calculate size }
         inc(textendaddr,textsize);

         { load data segment }
         datasize:=header.datalen;
         getmem(dataseg,datasize);
         blockread(f,textseg^,datasize);
         if ioresult<>0 then
           fatalerror(linker_file_read_error);
         { ... and calculate size }
         inc(dataendaddr,datasize);

         { setup bsssegment }
         bsssize:=header.bsslen;
         inc(bssendaddr,bsssize);

         { start address is ignored }

         { load relocs for text }
         textrelocs_count:=header.reloctext div 8;
         getmem(_textrelocs,header.reloctext);
         blockread(f,_textrelocs,header.reloctext);
         if ioresult<>0 then
           fatalerror(linker_file_read_error);

         { load relocs for data }
         datarelocs_count:=header.relocdata div 8;
         getmem(datarelocs,header.relocdata);
         blockread(f,_textrelocs,header.relocdata);
         if ioresult<>0 then
           fatalerror(linker_file_read_error);

         { load syms }
         getmem(syms,header.symslen);
         blockread(f,syms,header.symslen);
         if ioresult<>0 then
           fatalerror(linker_file_read_error);

         { load size of strings }
         blockread(f,l,4);
         if ioresult<>0 then
           fatalerror(linker_file_read_error);

         { load strings }
         getmem(strings,l-4);
         blockread(f,strings,l-4);
         if ioresult<>0 then
           fatalerror(linker_file_read_error);
         close(f);

         { now we have all needed informations }
         { step through all symbols }
         for l:=0 to (header.symslen div 12)-1 do
           begin
              hs:=new(plinksym,init(strings+syms^[l].name,syms^[l].value,syms^[l]._type and $fe));
              { global symbol ? }
              if (syms^[l]._type and 1)<>0 then
                begin
                   { if no extern definition then concat }
                   if syms^[l]._type<>1 then
                     insert(hs,global_syms);
                end
              else
                insert(hs,static_syms);
           end;
         for l:=0 to datarelocs_count-1 do
           begin

                number=0;
  datarelocsize/=8;
  while (datarelocsize>0)
  {
    trelocrecord *r=new trelocrecord;
    r->next=datarelocwurzel;
    datarelocwurzel=r;
    r->reloc_info=datarelocs[number];
    r->file=filedata;
    if (r->reloc_info.r_extern==1)
    {
      r->reloc_symbol=psymbol(symboltab[r->reloc_info.r_symbolnum].n_un.n_name);
    }
    datarelocsize--;
    number++;
  }

           end;
      end;

    procedure loadfile(const n : string);

      var
         magic : longint;
         f : file;
         ho : pobjfile;

      begin
         assign(f,n);
         {$I-}
         reset(f,1);
         if ioresult<>0 then
           begin
              exterror:=strpnew(n);
              error(linker_file_not_found);
              exit;
           end;
         blockread(f,magic,4);
         case magic and $ffff of
            $107 : begin
                      { it's a DOS o-file }
                      ho:=new(pdos_o_file,load(f));
                   end;
            else
               begin
                  exterror:=strpnew(n);
                  error(linker_illegal_magic);
               end;
         end;
         ho^.next:=ofiles;
         ofiles:=ho;
      end;

    procedure init;

      begin
         global_syms:=nil;
         ofiles:=nil;
      end;

end.
