{************************************************}
{  IFF Checker Version 1.53                      }
{  Copyright (C) Carl Eric Codere 1996           }
{  PROTECTED AND REAL MODES                      }
{************************************************}
Program IFFCheck;
{ + = Added feature                                                        }
{ $ = Bugfix                                                               }
{ - = Removed feature                                                      }
{ BUGFIXES/UPDATES                                                         }
{    + (10 March 1996)                                                     }
{          Added Nesting levels for FORMS and CATS. Maximum Nesting level  }
{           of IFF files that can be checked is MAXLEVELS (cf)             }
{    + (14 march 1996)                                                     }
{           Added screen pausing, requires CRT unit.                       }
{    + (6 june 1996)                                                       }
{           updated to work with iff39 versions.                           }
Uses Objects,CRT;

{*******************************************************************}
{                   PROGRAM CONSTANTS                               }
{*******************************************************************}
CONST
{ Basic group types }
 IdFORM    = 'FORM';
 IdPROP    = 'PROP';
 IdCAT     = 'CAT ';
 IdFILLER  = '    ';
 IdLIST    = 'LIST';
 Version   = '1.52';

 { Possible error codes returned to DOS by this program }
 EXIT_NOERROR   = 0;
 EXIT_NOTIFF    = 1;
 EXIT_DOSERROR  = 2;

 { Maximum possible nesting levels }
 MAXLEVELS = 4096;

 { Starting from ZERO! }
 MaxReservedID  = 30;
 MaxKnownForms  = 19;
 MaxKnownChunks = 8;

 { Reserved ID's }
 { THESE ID's ARE RESERVED AND CANNOT BE USED }
 ReservedID: Array[0..MaxReservedID] of String[4] =
 ('FORM','LIST','PROP','CAT ','FOR1','FOR2','FOR3','FOR4','FOR5','FOR6',
  'FOR7','FOR8','FOR9','LIS1','LIS2','LIS3','LIS4','LIS5','LIS6','LIS7',
  'LIS8','LIS9','CAT1','CAT2','CAT3','CAT4','CAT5','CAT6','CAT7','CAT8',
  'CAT9');
 { These ID's are already in use or are already defined }
 KnownForms: Array[0..MaxKnownForms] of String[4] =
 ('8SVX','ACBM','ANBM','ANIM','FNTR','FNTV','FTXT','GSCR','HEAD','ILBM',
  'PDEF','PGTB','PICS','PLBM','SMUS','USCR','UVOX','VDEOS','WORD','PBM ');
 KnownFormsStr: Array[0..MaxKnownForms] of String =
 ('8-bit sampled voice         ',
  'Amiga continuous bitmap     ',
  'Animated bitmap             ',
  'Animation format            ',
  'Raster font                 ',
  'Vector font                 ',
  'Formatted text              ',
  'General use musical score   ',
  'Flow wordprocessor format   ',
  'Interleaved bitmap image    ',
  'Deluxe print page definition',
  'Program traceback format    ',
  'Macintosh picture           ',
  'Interleaved bitmap image    ',
  'Simple musical score        ',
  'Uhuru sound musical score   ',
  'Uhuru sound macintosh voice ',
  'Deluxe video construction set',
  'Prowrite text file format   ',
  'Deluxe paint bitmap         ');
{ These are currently known id's for all IFF files  }
 KnownChunks: Array[0..MaxKnownChunks] of String[4] =
 ('ANNO', 'AUTH', 'CHRS','(c) ','CSET','FVER','NAME',
  'TEXT', 'BODY');

 KnownChunkStr: Array[0..MaxKnownChunks] of String =
 ('text annotations',
  'author''s name   ',
  '                ',
  'copyright notice',
  'character set   ',
  'file version    ',
  '                ',
  'simple ascii text',
  'main information ');







{ Converts a Motorola Longint to an Intel Longint and vice versa }
{ For word swapping use the Swap procedure included in the RTL   }
Function SwapLongint(L: Longint): Longint; 
Begin
 SwapLongInt := 5;
end;


TYPE
 { ID type definition }
 TID = Array[1..4] of Char;

 { Each Chunk starts with this information }
 PChunkHeader = ^TChunkHeader;
 TCHunkHeader = Record
  ChunkID: TID;                     { CHUNK ID         }
  ChunkSize: Longint;               { Size of Chunk    }
 end;

 { Each Group starts with this ID }
 PGroupHeader = ^TGroupHeader;
 TGroupHeader = Record
  GroupID: TID;   { PROP, CAR OR LIST }
  GroupSize: Longint;              { Size of Group     }
  SubGroupId: TID; { TYPE OF GROUP     }
 end;

 { Structure used for Nested FORM/CAT/LIST parser }
 PNestingInfo = ^TNestingInfo;
 TNestingInfo = Record
  GroupStart: Longint;          { Group Starting Position                 }
  GroupEnd: Longint;            { Group Ending position                   }
  Written: Boolean;             { Some Data has been written in this FORM }
 end;

 PIFFCheck = ^TIFFCheck;
 TIFFCheck =  Object(TObject)
 public
  { If ExtendedInfo is TRUE will right complete information }
  { about possible type of IFF file.                        }
  Constructor Init(AFname: FNameStr; Extended: Boolean);

  Destructor Done; Virtual;
  { Check the headers for consistency }
  Function CheckHeader(Header: TGroupHeader): Boolean;
  { Writes the output to the screen and also calls }

  { PrintGroupInfo if the specified option is set  }
  Procedure PrintGroup(Header: TGroupHeader; Nesting: Word);
  { Writes the Chunk information to the scren }
  Procedure PrintChunk(Chunk: TChunkHeader; Nesting: Word);
  { Checks the table of known groups, and compares with }
  { current value, if the same writes information to    }
  { screen.                                             }
  Procedure PrintGroupInfo(Str: String);
  { Checks the table of known chunks, and compares with }
  { current value, if the same, writes information to   }
  { screen.                                             }
  Procedure PrintChunkInfo(Str: TID);
  { Called when the number of lines on screen is over the limit }
  Procedure PauseScreen;
  { Main loop for analyzing the IFF files }
  Procedure Process;
 private

  ExtendedInfo: Boolean;
  CurrentID: TID;
  { Current Stream being processed }
  S: TBufStream;
  { Headers and information up to maxnesting levels deep }
  GroupHeader: Array[0..MAXLEVELS] of PGroupHeader;
  GroupInformation: Array[0..MAXLEVELS] of PNestingInfo;
  { Current Chunk being processed }
  ChunkHeader: TChunkHeader;
  { Current nesting level of IFF file }
  { 0 = means no level nesting        }
  NestingLevel: Word;
  { TOTAL number of chunks }
  ChunkLevels: Longint;
  { TOTAL number of groups }
  GroupLevels: Longint;
  { Number of lines on the screen }
  LinesOnScreen: Byte;
  { Checks if the entered name is valid }
  Function ValidName(Str: TID): Boolean;
  { Checks if the ID is already registered as a known ID }
  Function CheckChunk(Chunk: TChunkHeader): Boolean;
  Function ValidGroup(Str: string): Boolean;
 end;




  Function TIFFCheck.ValidGroup(Str: string): Boolean;
 {*******************************************************************}
 { TIFFFormWriter.ValidGroup(Str: String): Boolean;                  }
 {   returns TRUE if the ID's are valid, otherwise returns FALSE.    }
 {  Str -> Id to check.                                              }
 {*******************************************************************}
 VAR
  i: Byte;
 Begin
  ValidGroup := FALSE;
  { Check if the AChunk ID's are CAPITAL OR SPACE Characters }
  If (Str[1] in ['A'..'Z','1'..'9',' ']) AND (Str[2] in ['A'..'Z',' ','1'..'9']) AND
  (Str[3] in ['A'..'Z',' ','1'..'9']) AND (Str[4] in ['A'..'Z',' ','1'..'9']) then
   ValidGroup := TRUE;
  { Check if we are not using reserved words }
  For i:=0 to MaxReservedID do
  Begin
   If Str = ReservedID[i] then
   Begin
    ValidGroup := TRUE;
    Break;
   end;
  end;
end;


  Function TIFFCheck.ValidName(Str: TID): Boolean;
 {*******************************************************************}
 { TIFFFormWriter.ValidName(Str: array[1..4] of char;{ String }): Boolean;                   }
 {   Checks the validity of the different chunk and header ID's,     }
 {   returns TRUE if the ID's are valid, otherwise returns FALSE.    }
 {  Str -> Id to check.                                              }
 {*******************************************************************}
 VAR
  i: Byte;
 Begin
  ValidName := FALSE;
  { Check if the AChunk ID's are permitted Characters }
  {!!!!!!!
  If (Str[1] in [Char($20)..Char($7F)]) AND (Str[2] in [Char($20)..Char($7F)]) AND
  (Str[3] in [Char($20)..Char($7F)]) AND (Str[4] in [Char($20)..Char($7F)]) then
   ValidName := TRUE;
  }
  { Check if we are not using reserved words }
  For i:=0 to MaxReservedID do
  Begin
   { 
   If Str = ReservedID[i] then
   Begin
    ValidName := FALSE;
    Break;
   end;
   }
  end;
 end;



 Function TIFFCheck.CheckHeader(Header: TGroupHeader): Boolean;
 { Verifies if the ID HEADER is OK                }
 { Verifies, after swapping that the size of the  }
 { file, at least has this size.                  }
 { finally checks if the Sub-group ID is ok.      }
 { RETURNS TRUE if everything OK.                 }
 Begin
  CheckHeader := TRUE;
  { Check if the Group ID is ok }
  If NOT ValidGroup(Header.GroupID) then

  Begin
   CheckHeader := FALSE;
   Exit;

  end;
  { If the longint values are correct }
  If S.GetSize < Header.GroupSize then
  Begin
   CheckHeader := FALSE;
   Exit;
  end;
 If NOT ValidGroup(Header.SubGroupID) then
  Begin
   CheckHeader := FALSE;
   Exit;
  end;
 end;


   FUNCTION TIFFCheck.CheckChunk(Chunk: TChunkHeader): Boolean;

 {*******************************************************************}
 { - Verifies if the Chunk Header is correct                         }
 { - Verifies if the chunk size is correct                           }
 {                                                                   }
 {*******************************************************************}

  Begin
   CheckChunk := TRUE;
   If Not ValidName(Chunk.ChunkID) then
     CheckChunk := FALSE;
   If S.GetSize < (Chunk.ChunkSize) then
     CheckChunk := FALSE;
  end;



  Constructor TIFFCheck.Init(AFname: FNameStr; Extended: Boolean);
  Begin
   Inherited Init;
   ExtendedInfo := Extended;
   WriteLn('Opening IFF file.');
   {!!!!!
   S.Init(AFName, StOpenRead, 2048);
   }
   If S.GetSize <  Sizeof(TGroupHeader) then
   Begin
    WriteLn('Invalid File Size. Probably not an IFF file.');
    Halt(EXIT_NOTIFF);
   end;
   New(GroupHeader[NestingLevel]);
   New(GroupInformation[NestingLevel]);

   S.Read(GroupHeader[NestingLevel]^,  Sizeof(TGroupHeader));
   GroupHeader[NestingLevel]^.GroupSize := SwapLongint(GroupHeader[NestingLevel]^.GroupSize);
   GroupInformation[NestingLevel]^.GroupStart := S.GetPos;
   GroupInformation[NestingLevel]^.GroupEnd := S.GetPos + GroupHeader[NestingLevel]^.GroupSize - 8;
   If NOT CheckHeader(GroupHeader[NestingLevel]^) then
   Begin
    WriteLn('ERROR: Invalid IFF file');
    Halt(EXIT_NOTIFF);
   end;
   PrintGroup(GroupHeader[NestingLevel]^, NestingLevel);
   Inc(GroupLevels);
   Process;
  end;

  Procedure TIFFCheck.PauseScreen;
  { Pauses the screen }
  Var
   Ch: Char;
   XPos: Byte;
   YPos: Byte;
  Begin
   XPos := WhereX;
   YPos := WhereY;
   Write('-- Pause --');
   Repeat
    { Return key is pressed }
   Until Keypressed;
    Ch := Readkey;
   GotoXY(XPos, YPos);
   Write('            ');
   GotoXY(XPos, YPos);
  end;

  Procedure TIFFCheck.Process;
  Var
   i: Word;
   Pos: Longint;
  Begin
   Repeat
      { READ IN IDENTIFICATION TYPE   }
      S.Read(CurrentID, SizeOf(TID));
      { SEEK BACK THE FOUR CHARACTERS }
      S.Seek(S.GetPos - SizeOf(Tid));
      {!!!!!!
      If (CurrentID = IDForm) OR (CurrentID = IDCat) OR (CurrentID = IdList) then
      }
      if true then
      Begin
       Inc(GroupLevels);
       Inc(NestingLevel);
       New(GroupHeader[NestingLevel]);
       New(GroupInformation[NestingLevel]);
       S.Read(GroupHeader[NestingLevel]^,  Sizeof(TGroupHeader));
       GroupHeader[NestingLevel]^.GroupSize := SwapLongint(GroupHeader[NestingLevel]^.GroupSize);
       GroupInformation[NestingLevel]^.GroupStart := S.GetPos;
       GroupInformation[NestingLevel]^.GroupEnd := S.GetPos + GroupHeader[NestingLevel]^.GroupSize -4;
       If NOT CheckHeader(GroupHeader[NestingLevel]^) then
       Begin
        WriteLn('ERROR: Invalid IFF file');
        Halt(EXIT_NOTIFF);
       end;
       PrintGroup(GroupHeader[NestingLevel]^, NestingLevel);
      end
      else
      Begin
        S.Read(ChunkHeader, Sizeof(TChunkHeader));
        ChunkHeader.ChunkSize:= SwapLongint(ChunkHeader.ChunkSize);
        { Verify if everything is OK in this chunk }
        If NOT CheckChunk(ChunkHeader) then
        Begin
          WriteLn('ERROR: Invalid chunk: ',ChunkHeader.ChunkID);
          Halt(EXIT_NOTIFF);
        end;
        PrintChunk(ChunkHeader, NestingLevel+1);
        { Do not forget the odd bytes }
        If Odd(ChunkHeader.ChunkSize) then
           S.Seek(S.GetPos + ChunkHeader.ChunkSize + 1)
        else
           S.Seek(S.GetPos + ChunkHeader.ChunkSize);
        Inc(ChunkLevels);
     end;
      { Close all nesting levels when we have arrived at the end of the }
      { group                                                           }
      While S.GetPos = GroupInformation[NestingLevel]^.GroupEnd do
      Begin
       { Check for Error in nesting levels }
       If NestingLevel < 0 then
       Begin
        WriteLn('ERROR: Invalid nesting level at ',GroupHeader[NestingLevel]^.SubGroupID);
        Halt(EXIT_NOTIFF);
       end;
       Dispose(GroupHeader[NestingLevel]);
       Dispose(GroupInformation[NestingLevel]);
       Dec(NestingLevel);
     end;
  Until S.GetSize = S.GetPos;

 end;


  Procedure TIFFCheck.PrintGroup(Header: TGroupHeader; Nesting: Word);
  Var
   i: Word;
  Begin
  If Nesting <> 0 then
  Begin
     Nesting := Nesting SHL 1;
     For i:=0 to Nesting do
       Write(' ');
  end;
   WriteLn('Group -> ',Header.GroupID,' ',Header.SubGroupID);
   Inc(LinesOnScreen);
   For i:=1 to Nesting+1 do
    Write(' ');
   Write('Size of Group: ',Header.GroupSize);
   { Write information about current SubGroupID only }
   { if we requested maximum information             }
   If ExtendedInfo then
    PrintGroupInfo(Header.SubGroupID)
   else
    WriteLn;
   Inc(LinesOnScreen);
   { PrintGroupInfo is printed on same line as groupsize }
   If LinesOnScreen >= Hi(WindMax)-2 then
   Begin
    PauseScreen;
    LinesOnScreen := 0;
   end;
  END;


  Procedure TIFFCheck.PrintGroupInfo(Str:String);
  Var
   i: Integer;
   j: Integer;
  Begin
   For i:=0 to MaxKnownForms do
   { Iterate through table of known group ID's }
   { if found writeout appropriate string to   }
   { console.                                  }
   Begin
    If Str = KnownForms[i] then
    Begin
     { Write the information at an absolute value }
      For j := WhereX to 40 do
        Write(' ');
      WriteLn(KnownFormsStr[i]);
      { Once we have found it exit loop immediately and save some }
      { precious time.                                            }
      Exit;
   end;
   end;
   { Not known - writeout information about this }
   { at an absolute value from start of coordinate }
   For j := WhereX to 40 do
     Write(' ');
   WriteLn('Unknown group Identification');
  end;



  Procedure TIFFCheck.PrintChunk(Chunk: TChunkHeader; Nesting: Word);
  Var
   i: Word;
  Begin
  If Nesting <> 0 then
  Begin
    Nesting := Nesting SHL 1;
     For i:=0 to Nesting do
       Write(' ');
  end;
   Write('Chunk -> ',Chunk.ChunkID, '  ',Chunk.ChunkSize);
   { If ExtendedInfo is on, write out extended information on }
   { chunks (if available)                                    }
   If ExtendedInfo then
    PrintChunkInfo(Chunk.ChunkID)
   else
    WriteLn;
   Inc(LinesOnScreen);
   { Pause the screen }
   If LinesOnScreen >= Hi(WindMax)-2 then
   Begin
    PauseScreen;
    LinesOnScreen := 0;
   end;
  end;


  Procedure TIFFCheck.PrintChunkInfo(Str: TID);
{*******************************************************************}
{ This procedure prints out the chunk extended information to the   }
{ console, it goes through a table and matches the chunk names with }
{ their description, and then prints out the result to the console. }
{                                                                   }
{*******************************************************************}
  Var
   i: Integer;
   j: Integer;
  Begin
   For I:= 0 to MaxKnownChunks do
   Begin
    { If we have identified a chunk }
    { in whatever group             }
    {!!!!!
    If KnownChunks[i] = Str then

    Begin
     For j:=WhereX to 42 do
      Write(' ');
     WriteLn(KnownChunkStr[i]);
     Exit;
    end;
    }
   end;
    WriteLn;
  end;

  Destructor TIFFCheck.Done;
  Begin
   { Verify if all nesting levels have been closed }
   { if not there is an error in the IFF file      }
   If NestingLevel > 0 then
   Begin
    WriteLn('ERROR: Invalid group nesting levels - some groups not closed.');
    Halt(EXIT_NOTIFF);
   end;
   If Assigned(GroupHeader[NestingLevel]) then Dispose(GroupHeader[NestingLevel]);
   If Assigned(GroupInformation[NestingLevel]) then Dispose(GroupInformation[NestingLevel]);
   S.Done;
   WriteLn('Closing IFF file.');
   WriteLn('Report           ');
   WriteLn(' number of groups: ',GroupLevels);
   WriteLn(' number of chunks: ',ChunkLevels);
   Inherited Done;
  end;


{***********************************************************************}
{ Create a stream error procedure which will be called on error of the  }
{ stream. Will Terminate executing program, as well as display info     }
{ on the type of error encountered.                                     }
{***********************************************************************}
Procedure StreamErrorProcedure(Var S: TStream); FAR;

Begin
 If S.Status = StError then
 Begin
  WriteLn('ERROR: General Access failure. Halting');
  Halt(EXIT_DOSERROR);
 end;
 If S.Status = StInitError then
 Begin
  Write('ERROR: Cannot Init Stream. Halting. ');
  { SPECIFIC TO DOS STREAMS }
  Case S.ErrorInfo of
  2: WriteLn('File not found.');
  3: WriteLn('Path not found.');
  5: Writeln('Access denied.');
  else
    WriteLn;
  end;

  Halt(EXIT_DOSERROR);
 end;
 If S.Status = StReadError then
 Begin
  WriteLn('ERROR: Read beyond end of Stream. Halting');
  Halt(EXIT_DOSERROR);
 end;
 If S.Status = StWriteError then
 Begin
  WriteLn('ERROR: Cannot expand Stream. Halting');
  Halt(EXIT_DOSERROR);
 end;
 If S.Status = StGetError then
 Begin
  WriteLn('ERROR: Get of Unregistered type. Halting');
  Halt(EXIT_DOSERROR);

 end;
 If S.Status = StPutError then
 Begin
  WriteLn('ERROR: Put of Unregistered type. Halting');
  Halt(EXIT_DOSERROR);
 end;
end;

Procedure WriteInformation;
{ Writes information about the program }
Begin
   WriteLn('Simple IFF checker/identifier - Carl Eric Codere 1996');
   WriteLn('Freeware - Version ',Version);
   WriteLn;
   WriteLn('Usage: fname.ext to check');
   Halt(EXIT_NOERROR);
end;


VAR
 IFF: TIFFCheck;
Begin
  StreamError:= @StreamErrorProcedure;
  If Paramcount = 0 then WriteInformation;
  IFF.Init(ParamStr(1),TRUE);
  IFF.Done;
end.
